import pathlib

import pytest

from jschon import JSON, JSONSchema
from jschon.jsonpatch import JSONPatch
from jschon.utils import json_loadf

testsuite_dir = pathlib.Path(__file__).parent.parent / 'jschon_translation' / 'json-translation-vocabulary' / 'tests'


def pytest_generate_tests(metafunc):
    argnames = ('schema', 'data', 'patches', 'translations')
    argvalues = []
    testids = []

    testfile_paths = sorted(testsuite_dir.glob('*.json'))
    for testfile_path in testfile_paths:
        testcase = json_loadf(testfile_path)
        for n, test in enumerate(testcase['tests']):
            argvalues += [pytest.param(testcase['schema'], test['data'], test['patches'], test['translations'])]
            testids += [f'{testfile_path.name}/{n}']

    metafunc.parametrize(argnames, argvalues, ids=testids)


def test_translate_json(schema, data, patches, translations):
    schema = JSONSchema(schema)
    data = JSON(data)
    patches = {
        scheme: JSONPatch(*patch)
        for scheme, patch in patches.items()
    }

    # sanity checks
    assert schema.validate().valid is True
    assert patches.keys() == translations.keys()
    for scheme, patch in patches.items():
        assert patch.evaluate(None) == translations[scheme]

    result = schema.evaluate(data)
    assert result.valid
    for scheme, patch in patches.items():
        assert result.output('translation-patch', scheme=scheme) == patch
    for scheme, translation in translations.items():
        assert result.output('translation', scheme=scheme) == translation



import pathlib
from datetime import datetime
from urllib.parse import urlparse

from jschon import JSON, JSONSchema
from jschon.jsonpatch import JSONPatch
from jschon_translation import translation_filter

examples_dir = pathlib.Path(__file__).parent.parent / 'jschon_translation' / 'json-translation-vocabulary' / 'examples'


@translation_filter('date-to-year')
def date_to_year(date: str) -> int:
    return datetime.strptime(date, '%Y-%m-%d').year


@translation_filter('base-url')
def base_url(url: str) -> str:
    u = urlparse(url)
    return f'{u.scheme}://{u.netloc}'


def test_translate_iso19115_to_datacite():
    example_dir = examples_dir / 'iso19115-to-datacite'
    input_schema = JSONSchema.loadf(example_dir / 'input-schema.json')
    input_json = JSON.loadf(example_dir / 'input.json')
    output_schema = JSONSchema.loadf(example_dir / 'output-schema.json')
    output_json = JSON.loadf(example_dir / 'output.json')

    # sanity checks
    assert input_schema.validate().valid
    assert input_schema.evaluate(input_json).valid
    assert output_schema.validate().valid
    assert output_schema.evaluate(output_json).valid

    result = input_schema.evaluate(input_json)
    patch = result.output('translation-patch', scheme='datacite')
    translation = result.output('translation', scheme='datacite', clear_empties=False)
    assert JSONPatch(*patch).evaluate(None) == translation
    translation = result.output('translation', scheme='datacite', clear_empties=True)
    assert translation == output_json



from typing import Mapping, Optional, Sequence

from jschon.exc import JSONSchemaError
from jschon.json import JSON, JSONCompatible
from jschon.jsonpointer import JSONPointer, RelativeJSONPointer
from jschon.jsonschema import JSONSchema, Result
from jschon.vocabulary import Keyword, Subschema, SubschemaMixin
from jschon_translation import JSONTranslationSchema, TranslationResult

__all__ = [
    'TranslationsKeyword',
    'T9nSchemeKeyword',
    'T9nTargetKeyword',
    'T9nConditionKeyword',
    'T9nConstKeyword',
    'T9nSourceKeyword',
    'T9nConcatKeyword',
    'T9nSepKeyword',
    'T9nFilterKeyword',
    'T9nCastKeyword',
    'T9nArrayKeyword',
    'T9nObjectKeyword',
]


class TranslationsKeyword(Keyword, SubschemaMixin):
    key = "translations"

    @classmethod
    def jsonify(
            cls,
            parentschema: JSONSchema,
            key: str,
            value: Sequence[Mapping[str, JSONCompatible]],
    ) -> Optional[JSON]:
        return JSON(
            value,
            parent=parentschema,
            key=key,
            itemclass=JSONTranslationSchema,
            catalog=parentschema.catalog,
            cacheid=parentschema.cacheid,
        )

    def evaluate(self, instance: JSON, result: Result) -> None:
        for index, subschema in enumerate(self.json):
            with result(instance, str(index), cls=TranslationResult) as subresult:
                subschema.evaluate(instance, subresult)


class T9nSchemeKeyword(Keyword):
    key = "t9nScheme"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: str):
        super().__init__(parentschema, value)
        parentschema.t9n_scheme = value


class T9nTargetKeyword(Keyword):
    key = "t9nTarget"

    def evaluate(self, instance: JSON, result: TranslationResult) -> None:
        result.parent.t9n_target = JSONPointer(self.json.value)


class T9nConditionKeyword(Keyword, Subschema):
    key = "t9nCondition"

    def evaluate(self, instance: JSON, result: TranslationResult) -> None:
        self.json.evaluate(instance, result)
        if not result.valid:
            result.parent.fail()


class T9nConstKeyword(Keyword):
    key = "t9nConst"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: JSONCompatible):
        super().__init__(parentschema, value)
        parentschema.t9n_const = value


class T9nSourceKeyword(Keyword):
    key = "t9nSource"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: str):
        super().__init__(parentschema, value)
        if (source := RelativeJSONPointer(value)).index:
            raise JSONSchemaError('"t9nSource" must reference an instance, not an index')
        parentschema.t9n_source = source


class T9nConcatKeyword(Keyword):
    key = "t9nConcat"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: Sequence[str]):
        super().__init__(parentschema, value)
        parentschema.t9n_concat = tuple(RelativeJSONPointer(v) for v in value)


class T9nSepKeyword(Keyword):
    key = "t9nSep"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: str):
        super().__init__(parentschema, value)
        parentschema.t9n_sep = value


class T9nFilterKeyword(Keyword):
    key = "t9nFilter"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: str):
        super().__init__(parentschema, value)
        parentschema.t9n_filter = value


class T9nCastKeyword(Keyword):
    key = "t9nCast"
    static = True

    def __init__(self, parentschema: JSONTranslationSchema, value: str):
        super().__init__(parentschema, value)
        parentschema.t9n_cast = value


class T9nArrayKeyword(Keyword, SubschemaMixin):
    key = "t9nArray"
    depends_on = "t9nCondition",

    def __init__(self, parentschema: JSONTranslationSchema, value: Mapping[str, JSONCompatible]):
        super().__init__(parentschema, value)
        parentschema.t9n_leaf = False

    @classmethod
    def jsonify(
            cls,
            parentschema: JSONTranslationSchema,
            key: str,
            value: Mapping[str, JSONCompatible],
    ) -> Optional[JSON]:
        return JSONTranslationSchema(
            value,
            parent=parentschema,
            key=key,
            catalog=parentschema.catalog,
            cacheid=parentschema.cacheid,
            scheme=parentschema.t9n_scheme,
        )

    def evaluate(self, instance: JSON, result: TranslationResult) -> None:
        if (condition := result.sibling(instance, "t9nCondition")) and not condition.valid:
            return

        scheme = self.parentschema.t9n_scheme
        array_path = result.parent.t9n_target
        result.init_array(scheme, array_path)

        if instance.type == 'array':
            for item in instance:
                index = result.next_array_index(scheme, array_path)
                result.t9n_target = array_path / str(index)
                self.json.evaluate(item, result)
        else:
            index = result.next_array_index(scheme, array_path)
            result.t9n_target = array_path / str(index)
            self.json.evaluate(instance, result)


class T9nObjectKeyword(Keyword, SubschemaMixin):
    key = "t9nObject"
    depends_on = "t9nCondition",

    def __init__(self, parentschema: JSONTranslationSchema, value: Mapping[str, JSONCompatible]):
        super().__init__(parentschema, value)
        parentschema.t9n_leaf = False

    @classmethod
    def jsonify(
            cls,
            parentschema: JSONTranslationSchema,
            key: str,
            value: Mapping[str, JSONCompatible],
    ) -> Optional[JSON]:
        return JSON(
            value,
            parent=parentschema,
            key=key,
            itemclass=JSONTranslationSchema,
            catalog=parentschema.catalog,
            cacheid=parentschema.cacheid,
            scheme=parentschema.t9n_scheme,
        )

    def evaluate(self, instance: JSON, result: TranslationResult) -> None:
        if (condition := result.sibling(instance, "t9nCondition")) and not condition.valid:
            return

        scheme = self.parentschema.t9n_scheme
        object_path = result.parent.t9n_target
        result.init_object(scheme, object_path)

        for name, subschema in self.json.items():
            with result(instance, name) as subresult:
                subresult.t9n_target = object_path / name
                subschema.evaluate(instance, subresult)



from __future__ import annotations

from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from jschon.exc import RelativeJSONPointerReferenceError
from jschon.json import JSON, JSONCompatible
from jschon.jsonpatch import JSONPatch, JSONPatchOperation, PatchOp
from jschon.jsonpointer import JSONPointer, RelativeJSONPointer
from jschon.jsonschema import JSONSchema, Result
from jschon.output import output_formatter

__all__ = [
    'JSONTranslationSchema',
    'TranslationResult',
    'TranslationFilter',
    'translation_filter',
]


class NoValue:
    pass


class JSONTranslationSchema(JSONSchema):

    def __init__(
            self,
            *args: Any,
            scheme: str = None,
            **kwargs: Any,
    ) -> None:
        self.t9n_scheme: Optional[str] = scheme
        self.t9n_source: Optional[RelativeJSONPointer] = None
        self.t9n_const: Optional[JSONCompatible] = NoValue
        self.t9n_concat: Optional[Tuple[RelativeJSONPointer, ...]] = None
        self.t9n_sep: str = ''
        self.t9n_filter: Optional[Union[str, Dict[str, JSONCompatible]]] = None
        self.t9n_cast: Optional[str] = None
        self.t9n_leaf: bool = True
        super().__init__(*args, **kwargs)

    def evaluate(self, instance: JSON, result: TranslationResult = None) -> Result:
        if self.t9n_source is not None:
            try:
                source = self.t9n_source.evaluate(instance)
            except RelativeJSONPointerReferenceError:
                return result
        else:
            source = instance

        super().evaluate(source, result)

        if result.valid and self.t9n_leaf:
            if self.t9n_const is not NoValue:
                value = self.t9n_const
            elif self.t9n_concat is not None:
                value = []
                for item in self.t9n_concat:
                    try:
                        value += [self._make_value(item.evaluate(source))]
                    except RelativeJSONPointerReferenceError:
                        pass
                if value:
                    value = self.t9n_sep.join(str(v) for v in value)
                else:
                    value = NoValue
            else:
                value = self._make_value(source)

            if value is not NoValue:
                result.add_translation_patch(self.t9n_scheme, result.t9n_target, value)

        return result

    def _make_value(self, instance: JSON) -> JSONCompatible:
        value = instance.value

        if isinstance(self.t9n_filter, str):
            if filter_fn := _translation_filters.get(self.t9n_filter):
                value = filter_fn(value)
        elif isinstance(self.t9n_filter, dict):
            value = self.t9n_filter.get(value, value)

        if self.t9n_cast == 'boolean':
            value = bool(value)
        elif self.t9n_cast == 'integer':
            value = int(value)
        elif self.t9n_cast == 'number':
            value = float(value)
        elif self.t9n_cast == 'string':
            value = str(value)

        return value


TranslationFilter = Callable[[JSONCompatible], JSONCompatible]
_translation_filters: Dict[str, TranslationFilter] = {}


def translation_filter(name: str = None):
    def decorator(f):
        filter_name = name if isinstance(name, str) else f.__name__
        _translation_filters[filter_name] = f
        return f

    return decorator(name) if callable(name) else decorator


class TranslationResult(Result):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.t9n_target: Optional[JSONPointer] = None
        self.t9n_patchops: Optional[Dict[str, List[JSONPatchOperation]]] = None

    def add_translation_patch(
            self,
            scheme: str,
            target: JSONPointer,
            value: JSONCompatible,
    ) -> None:
        if self.t9n_patchops is None:
            self.t9n_patchops = {}
        self.t9n_patchops.setdefault(scheme, [])
        self.t9n_patchops[scheme] += [JSONPatchOperation(
            op=PatchOp.ADD,
            path=target,
            value=value,
        )]

    def init_array(self, scheme: str, array_path: JSONPointer) -> None:
        self.globals.setdefault(scheme, {})
        self.globals[scheme].setdefault('arrays', {})
        if array_path not in self.globals[scheme]['arrays']:
            self.globals[scheme]['arrays'][array_path] = 0
            self.add_translation_patch(scheme, array_path, [])

    def next_array_index(self, scheme: str, array_path: JSONPointer) -> int:
        next_index = self.globals[scheme]['arrays'][array_path]
        self.globals[scheme]['arrays'][array_path] += 1
        return next_index

    def init_object(self, scheme: str, object_path: JSONPointer) -> None:
        self.globals.setdefault(scheme, {})
        self.globals[scheme].setdefault('objects', set())
        if object_path not in self.globals[scheme]['objects']:
            self.globals[scheme]['objects'] |= {object_path}
            self.add_translation_patch(scheme, object_path, {})


@output_formatter('translation-patch')
def translation_patch(
        result: Result,
        scheme: str,
        ignore_validity: bool = False,
) -> JSONCompatible:
    """Output formatter for the ``translation-patch`` output format.

    :param result: Evaluation result.
    :param scheme: Translation scheme.
    :param ignore_validity: True to ignore the validity of the result.
    :return: A translation patch (list of JSON patch operation dicts).
    """
    return JSONPatch(*_visit(result, scheme, ignore_validity)).aslist()


@output_formatter('translation')
def translation(
        result: Result,
        scheme: str,
        ignore_validity: bool = False,
        clear_empties: bool = False,
) -> JSONCompatible:
    """Output formatter for the ``translation`` output format.

    :param result: Evaluation result.
    :param scheme: Translation scheme.
    :param ignore_validity: True to ignore the validity of the result.
    :param clear_empties: True to remove all empty strings, lists, dicts,
        and nulls from the translated document.
    :return: The translated document.
    """
    output = JSONPatch(*_visit(result, scheme, ignore_validity)).evaluate(None)
    if clear_empties:
        _clear_empties(output)
    return output


def _visit(node: Result, scheme: str, ignore_validity: bool) -> Iterator[JSONPatchOperation]:
    if ignore_validity or node.valid:
        if hasattr(node, 't9n_patchops'):
            try:
                yield from node.t9n_patchops[scheme]
            except (KeyError, TypeError):
                pass
        for child in node.children.values():
            yield from _visit(child, scheme, ignore_validity)


def _clear_empties(node: JSONCompatible):
    if isinstance(node, list):
        for subnode in list(node):
            _clear_empties(subnode)
            if isinstance(subnode, (type(None), str, list, dict)) and not subnode:
                node.remove(subnode)
    elif isinstance(node, dict):
        for key in dict(node):
            _clear_empties(subnode := node[key])
            if isinstance(subnode, (type(None), str, list, dict)) and not subnode:
                del node[key]
